/****************************************************************
 *
 * smsPLCServ
 * Author:  Damon Spayde (spayde@jlab.org)
 * Time-stamp: <2003-08-11 13:56:44 spayde>
 * 
 ***************************************************************/

/****************************************************************
 * 
 ***************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>
#include <sys/time.h>
#include "smsServCommon.h"
#include "../ethersdk/include/defs.h"
#include "../ethersdk/include/hei.h"
#include <netinet/in.h>
#include <arpa/inet.h>

/****************************************************************
 *** Constant Declarations
 ***************************************************************/
//#define kDisablePLC 1  /* Disable PLC access routines */
#define kDisablePLC 0  /* Enable PLC access routines */

/****************************************************************
 *** Function Declarations
 ***************************************************************/
int get_options(int argc, char* argv[]);
int plc_initialization(const int n_monitors_ex);
int plc_poll();
int plc_cleanup();
DWORD timeGetTime();
int CCMAck(HEIDevice *pDevice);
int CCMResponse(HEIDevice *pDevice, BYTE *pData, WORD DataLen);

/****************************************************************
 *** Global Variable Declarations
 ***************************************************************/
char* ini_filename; /* Name of initalization file */
char* pipe_filename; /* Name of pipe file */
char monitor_id[kNumMonitors][kMaxLength]; /* Array of monitor names
					      to process */
char* monitor_ex[kNumMonitors]; /* Array of monitor names to
				   exclude */
int n_monitors;  /* Number of monitors in monitor_id */
char hardware_address[kMaxLength]; /* MAC address of PLC */
int min_ccm_msec; /* minimum time (in msec) to wait between writes to
		     PLC */
int keep_alive_address; /* memory address of keep alive bit */
int keep_alive_bit_number; /* bit of keep alive toggle */
FILE* in_file;   /* file handle for input pipe */
struct timeval write_wait; /* Time to wait between
			       PLC writes */
HEITransport TP;            /* Variable containing details of
			       transport for HEI devices */
HEIDevice aDevices[kMaxDevices]; /* Array to contain details of HEI
				    devices */
WORD DeviceCount;           /* Number of devices stored in aDevices */
enum CCMAckCode{ cacWaiting, cacAck, cacNak };
DWORD time_starttime;

/****************************************************************
 ************************ Main Program **************************
 ***************************************************************/
int main(int argc, char *argv[]) {

    /*
     * Process the command line
     */
    int ret = 0;
    ret = get_options(argc, argv);
    if ( ret < 0 ) {
	fprintf(stderr, "**smsPLCServ => Error in get_options().  Exiting . . .\n");
	return 1;
    }
    int n_monitors_ex = ret;

    /*
     * Initialization
     */
    ret = plc_initialization(n_monitors_ex);
    if ( ret != 0 ) {
	fprintf(stderr, "**smsPLCServ => Error in plc_initialization().  Exiting . . .\n");
	return 1;
    }

    printf("PLC TempService.ini Settings:  \n");
    printf("  H4-ECOM MAC Address = %s\n", hardware_address);
    printf("  Minimum wait time between PLC updates (msec) = %d\n",
	   min_ccm_msec);
    printf("  Keep Alive Address = %d\n", keep_alive_address);
    printf("  Keep Alive Bit Number = %d\n", keep_alive_bit_number);

    /*
     * Poll the PLC indefinitely
     */
    while (1) {
	ret = plc_poll();
	if ( ret != 0 ) {
	    fprintf(stderr, "**smsPLCServ => Error in plc_poll().  Exiting . . .\n");
	    return 1;
	}
	break;
    }

    fclose(in_file);
    in_file = NULL;

    plc_cleanup();

    return 0;
}

/***************************************************************/
/***************************************************************/
/*********************** Functions *****************************/
/***************************************************************/
/***************************************************************/

/****************************************************************
 * This function parses the command line and insures all necessary
 * information is there
 * Returns -1 on error
 ***************************************************************/
int get_options(int argc, char* argv[]) {
    ini_filename = NULL;
    pipe_filename = NULL;
    int c;

    while ((c = getopt(argc, argv, "p:i:")) != -1) {
	switch (c) {
	case 'p':
	    /* Pipe File Name */
	    pipe_filename = optarg;
	    break;
	case 'i':
	    /* Initialization File Name */
	    ini_filename = optarg;
	    break;
	case '?':
	    if (isprint(optopt)) {
		fprintf(stderr, "Unknown option `-%c`.\n", optopt);
	    } else {
		fprintf(stderr, "Unknown option character `%x`.\n", optopt);
	    }
	    return -1;
	default:
	    return -1;
	}
    }

    /* Must specify all three on command line */
    if (ini_filename == NULL || pipe_filename == NULL) {
	fprintf(stderr, 
		"usage:  smsPLCServ -i <ini filename> -p <pipe filename> \n");
	return -1;
    }

    return 0;
}

/****************************************************************
 * This function performs all initialization tasks necessary for
 * forming a link with the PLC.
 * Returns 0, else nonzero on error
 ***************************************************************/
int plc_initialization(const int n_monitors_ex) {

//    n_monitors = 0;
//    int i = 0;
//    for (i = 0; i < kNumMonitors; i++) {
//	monitor_file[i] = NULL;
//    }
    in_file = NULL;

    DeviceCount = kMaxDevices;
    memset(&TP, 0, sizeof(TP));
    memset(aDevices, 0, sizeof(aDevices));

    /*
     * Open a connection to the PLC
     */
    if (!kDisablePLC) {
	HEIOpen(HEIAPIVERSION);
	TP.Transport = HEIT_UNIX;
	TP.Protocol = HEIP_IP;

	ENetAddress* hostAddr;
	hostAddr = malloc(sizeof(ENetAddress));
	memset(hostAddr, 0, sizeof(*hostAddr));
	TP.pSourceAddress = hostAddr;
	TP.pSourceAddress->AddressIP.Family = AF_INET;
	TP.pSourceAddress->AddressIP.Port = kHEIPort;
	//	TP.pSourceAddress->AddressIP.Port = 0x7070;
	struct in_addr tempAddr;
	if (!inet_aton(kHEIInterface, &tempAddr)) {
	    fprintf(stderr, "**smsPLCServ => Invalid host address!\n");
	    return -1;
	}
	TP.pSourceAddress->AddressIP.AddressingType.lAddr = tempAddr.s_addr;

	printf("DeviceCount = %d\n", DeviceCount);
	if (!HEIOpenTransport(&TP, HEIAPIVERSION, 0)) {
	    printf("PLC Server:  Successfully opened transport!\n");
	    HEIQueryDevices(&TP, aDevices, &DeviceCount, HEIAPIVERSION);
	    printf("PLC Server:  Detected %d device(s) on network interface %s.\n", 
		   DeviceCount, kHEIInterface);
	    if (DeviceCount <= 0) {
		/* Detected no devices on the network */
		fprintf(stderr, "**plc_initialization => Unable to locate any PLC devices!  Exiting . . .\n");
		return(1);
	    } else {
		/* Detected at least one device so open a connection */
		if (!HEIOpenDevice(&TP, &aDevices[0], HEIAPIVERSION, 
				   25, 3, FALSE)) {
		    /* Successfully opened a connection */
		    ;
		} else {
		    /* Failed to open a connection */
		    fprintf(stderr, "**plc_initialization => Unable to open device %02X%02X%02X%02X%02X%02X\n", aDevices[0].ENetAddress[0], 
			    aDevices[0].ENetAddress[1], 
			    aDevices[0].ENetAddress[2],
			    aDevices[0].ENetAddress[3],
			    aDevices[0].ENetAddress[4],
			    aDevices[0].ENetAddress[5]);
		    return(1);
		}
	    }
	} else {
	    fprintf(stderr, 
		    "**plc_initialization => Failed to open transport!\n");
	    return(1);
	}
    }
	

    /* Open ini file for reading */
    FILE* ini_file = NULL;
    ini_file = fopen(ini_filename, "r");
    if ( !ini_file ) {
	fprintf(stderr, "**plc_initialization => Unable to open %s.\n", 
		ini_filename);
	return 1;
    }

    /* 
     * Parse PLC section of file 
     */
    const int length = 100;
    char* ini_line = NULL;
    ini_line = (char*) malloc(length);
    strncpy(ini_line, "\0", length);
    char* ini_line_save = ini_line;
    int size = 10000;
    char* char_ptr = NULL;

    /* Locate beginning of PLC section */
    int ret = find_section(ini_file, "PLC");
    if (ret != 0) {
	exit(1);
    }

    while (1) { /* Loop over ini_file */

	ini_line = ini_line_save;
	size = getline(ini_line, length, ini_file);

	/* Check for end of file or of section */
	if ( feof(ini_file) || ini_line[0]==kSectionOpen ) {
	    break;
	}

	/* Check for valid entries and parse */
	if ((char_ptr = key_value(ini_line, "HardwareAddress"))) {
	    strcpy(hardware_address, char_ptr);
	} else if ((char_ptr = key_value(ini_line, "MinCCMmSec"))) {
	    min_ccm_msec = atoi(char_ptr);
	    write_wait.tv_sec = 0;
	    write_wait.tv_usec = min_ccm_msec * 1000;
	} else if ((char_ptr = key_value(ini_line, "KeepAliveAddress"))) {
	    keep_alive_address = atoi(char_ptr);
	} else if ((char_ptr = key_value(ini_line, "KeepAliveBitNumber"))) {
	    keep_alive_bit_number = atoi(char_ptr);
	} else {
	    ;
	}

    } /* End loop over ini_file */

    /* Close ini file */
    fclose(ini_file);

    /* Check to make sure input pipe exists and is readable */
    int err = access(pipe_filename, F_OK);
    if (err != 0) {
	perror("plc_initialization()");
	return -1;
    }

    err = access(pipe_filename, R_OK);
    if (err != 0) {
	perror("plc_initialization()");
	return -1;
    }
    printf("PLC Server:  File %s exists and is readable . . .\n", 
	   pipe_filename);
    
    /* Open input pipe */
    in_file = fopen(pipe_filename, "r");
    if (in_file == NULL) {
	fprintf(stderr, 
		"**plc_initialization => Unable to open %s . . .\n",
		pipe_filename);
	return -1;
    }
    printf("PLC Server:  Successfully opened %s for writing . . .\n", 
	   pipe_filename);

    /* Clean up */
    free(ini_line); ini_line = NULL;

    return 0;
}

/****************************************************************
 * This function reads temperature data from files and sends it to the
 * PLC.
 * Returns 0, else nonzero on error
 ***************************************************************/
int plc_poll() {

    int size = 10000;
    int length = 50;
    char mon_line[length];
    unsigned int plc_address = 0;
    float temp_value = 0;
    unsigned int* temp_value_raw;
    unsigned int monitor = 0;
    unsigned int channel = 0;
    unsigned int monitor_buffer[kNumMonitors][kNumChannels][2];
    struct timeval current_time, last_update_time, elapsed_time, diff_time;

    int err = gettimeofday(&last_update_time, NULL);
    if (err != 0) {
	perror("plc_poll()");
	return -1;
    }
    err = gettimeofday(&current_time, NULL);
    if (err != 0) {
	perror("plc_poll()");
	return -1;
    }

    /* Initialize data holding buffer to zero */
    int i, j, k;
    for (i = 0; i < kNumMonitors; i++) {
	for (j = 0; j < kNumChannels; j++) {
	    for (k = 0; k < 2; k++) {
		monitor_buffer[i][j][k] = 0;
	    }
	}
    }

    BOOL bWrite = TRUE;   /* Writing to PLC */
    BYTE DataType = 0x31; /* V memory is data type 0x31 */
    WORD Address = 0;     /* Memory address (CCM = PLC+1 in hex) */
    WORD DataLen = 0;     /* Amount of data to transmit (in bytes) */
    BYTE pData[100];       /* Data buffer */
    memcpy(pData, "0", sizeof(pData));
    
    while (1) { /* Loop over file */

	/* For each monitor read in each line and send to PLC until
	   EOF or kNumChannels read */
	size = getline(mon_line, length, in_file);
	if ( feof(in_file) ) {
	    break;
	}	
	
	/* Parsing data line into component pieces:
	   1st column = PLC address in octal
	   2nd column = temperature to write into PLC (float
	   format)
	*/
	sscanf(mon_line, "%u %u %o %f", &monitor, &channel, 
	       &plc_address, &temp_value);
	
	/* By the following trick we get access to the underlying
	   representation of the float.  The result is suitable
	   for transmission to the PLC */
	temp_value_raw = &temp_value;
	//	printf("**plc_poll => %d %d %o %f %x\n", monitor, channel, plc_address, temp_value, *temp_value_raw);
	
	/* CCM protocol adds one to the PLC address and transports
	   it as a hex number */
	plc_address++;
	
	/* Store data in monitor_buffer */
	monitor_buffer[monitor-1][channel-1][0] = plc_address;
	monitor_buffer[monitor-1][channel-1][1] = *temp_value_raw;

	/*
	printf("**plc_poll() => %d %d %o %d %f\n", monitor, channel,
	       monitor_buffer[monitor-1][channel-1][0],
	       monitor_buffer[monitor-1][channel-1][1],
	       temp_value);
	*/

	/* Check to see if enough time has elapsed since last
	   transmission to PLC to transmit again */
	err = gettimeofday(&current_time, NULL);
	if (err != 0) {
	    perror("plc_poll()");
	    return -1;
	}
	int time_sign = timeval_subtract(&elapsed_time, current_time, 
					 last_update_time);
	if (time_sign != 0) {
	    fprintf(stderr, 
		    "**plc_poll():  timeval_subtract returned %d . . .\n", 
		    time_sign);
	}

	/* time_sign should be negative if elapsed time is greater
	   than the write_wait time */
	time_sign = timeval_subtract(&diff_time, write_wait, elapsed_time);

	if (time_sign > 0) {
	    /* Enough time has elapsed so that the PLC can be updated
	       */
	    //	    printf("**plc_poll() => diff_time.tv_sec = %d, diff_time.tv_usec = %d, time_sign = %d\n", 
	    //		   diff_time.tv_sec, diff_time.tv_usec, time_sign);
	    for (i = 0; i < kNumMonitors; i++) {
		for (j = 0; j < kNumChannels; j++) {
		    /* Check to make sure that address is non-zero
		       before proceeding. */
		    if (monitor_buffer[i][j][0] == 0) {
			continue;
		    }
		    Address = monitor_buffer[i][j][0];
		    DataLen = sizeof(monitor_buffer[i][j][1]);
		    //		    sprintf(pData, "%x", monitor_buffer[i][j][1]);
		    /*
		    pData[3] = 12;
		    pData[2] = 34;
		    pData[1] = 56;
		    pData[0] = 78;
		    */
		
		    unsigned int m_buff = monitor_buffer[i][j][1];
		    pData[0] = m_buff & ~0xffffff00;
		    m_buff = m_buff >> 8;
		    pData[1] = m_buff & ~0xffffff00;
		    m_buff = m_buff >> 8;
		    pData[2] = m_buff & ~0xffffff00;
		    m_buff = m_buff >> 8;
		    pData[3] = m_buff & ~0xffffff00;
		    /*
		    pData[1] = m_buff & ~0xffff00ff;
		    pData[2] = m_buff & ~0xff00ffff;
		    pData[3] = m_buff & ~0x00ffffff;
		    */

		    //		    sprintf(pData, "12345678");
		    //		    DataLen = sizeof(pData);

		    /*
		    printf("**plc_poll()=> %d %d %o %x %d %x %x %x %x\n", 
			   i+1, j+1, Address, Address, DataLen, 
			   pData[0], pData[1], pData[2], pData[3]
			   );
		    */
		    
		    if (!kDisablePLC) {
			/* Set to non-blocking mode */
			aDevices[0].ParallelPackets = TRUE;
			
			/* Send CCM request */
			DWORD StartTime = timeGetTime();
			BOOL Done = FALSE, Acked = FALSE;
			int Error = HEICCMRequest(&aDevices[0], bWrite, 
						  DataType, Address, 
						  DataLen, pData);
	   
			if (Error == HEIE_NO_RESPONSE) {
			    /* Get CCM ack first */
			    do {
				switch(CCMAck(&aDevices[0])) {
				case cacWaiting:
				    /* Keep waiting */
				    break;
				case cacNak:
				    /* Nak can occur when module queues are full
				     */
				    fprintf(stderr, "**plc_poll => CCM request was NAKed.\n");
				    Done = TRUE;
				    break;
				case cacAck:
				    /* Received a response */
				    StartTime = timeGetTime();
				    Done = TRUE;
				    Acked = TRUE;
				    break;
				default:
				    /* Should never get here */
				    Done = TRUE;
				    fprintf(stderr, 
					    "**plc_poll => Unable to get CCM response.\n");
				    break;
				}
				/* Timeout */
			    } while(!Done && (timeGetTime() - StartTime < 25));
			    
			    Done = FALSE;
			    
			    /* Get response if request was ACKed */
			    if (Acked) {
				do {
				    switch(CCMResponse(&aDevices[0], 
						       pData, DataLen)) {
				    case HEIE_NO_RESPONSE:
					/* Wait */
					break;
				    case HEIE_NULL:
					/* */
					Done = TRUE;
					break;
				    default:
					Done = TRUE;
					fprintf(stderr, 
						"**plc_poll => Unable to get CCM response.\n");
					break;
				    }
				    /* 250 ms timeout.  Actual should
				       be max PLC scan time *2 + 5 to
				       10 ms */
				} while (!Done && 
					 (timeGetTime() - StartTime < 250));
				
				if (!Done)
				    fprintf(stderr, 
					    "**plc_poll => Timeout waiting for response..\n");
			    }
			} else {
			    fprintf(stderr, 
				    "**plc_poll => Unable to send CCM request.\n");
			}
		    } /* End kDisablePLC check */
		} /* End loop over channels */
	    } /* End loop over monitors */
	    err = gettimeofday(&last_update_time, NULL);
	    if (err != 0) {
		perror("plc_poll()");
		return -1;
	    }
	    //	    return -1; /*!!!!!!! REMOVE !!!!!!!*/
	} /* End elapsed time check */
    } /* End loop over file */
    return 0;
}

/****************************************************************
 * This function cleans up and shuts down the interface.
 * Returns 0, else nonzero on error
 ***************************************************************/
int plc_cleanup() {

    HEICloseDevice(&aDevices[0]);
    HEICloseTransport(&TP);
    HEIClose();

    return(0);

} /* End plc_cleanup() */

/* We need to supply a timeGetTime function to provide the timeouts
   that windows doesn't provide */
DWORD timeGetTime()
{
  /* We will use the gettimeofday function to provide the millisecond
     tick counter.  We will return the timer tick relative to the
     first time the routine is called    */
         struct timeval date;
         struct timezone tz;
         unsigned long now;
         gettimeofday(&date,&tz);
         /* The first time this timer is called, initialize the starttime
	 ** variable */
         if (time_starttime==0) {time_starttime = date.tv_sec;};
         now = ((date.tv_sec-time_starttime)*1000)+(date.tv_usec/1000);
         return now;
}



/*
** Note:
** Much of this in here is magic. The non-blocking mechinism isn't
** well defined yet. If you want to do this email and we'll give more
** info. We will get all of this documented as soon as possible.
** rmo
*/
int CCMAck(HEIDevice *pDevice)
{
	// Temp buffers for response
	BYTE Buff[500];
	int BuffSize = sizeof(Buff);

	// Check for a response
	int Error = HEIGetResponse(pDevice, Buff, &BuffSize, TRUE);
	
	// If error, deal with it
	if(Error)
		{
		if(Error == HEIE_NO_RESPONSE)
			return cacWaiting;
		else
			return Error;
		}
	// No error, test response
	else
		{
		if(Buff[PACKET_HEADER_SIZE] == FUN_ACK)
			return cacAck;
		else
			return cacNak;
		}
}

int CCMResponse(HEIDevice *pDevice, BYTE *pData, WORD DataLen)
{
	// Temp buffers for response
	BYTE Buff[500];
	int BuffSize = sizeof(Buff);

	// Check for a response
	int Error = HEIGetResponse(pDevice, Buff, &BuffSize, TRUE);

	// If error, return it
	if(Error)
		return Error;

	// Check for error from device
	Error = *(short int *)(Buff+PACKET_HEADER_SIZE+2);
	
	// If error, return it
	if(Error)
		return Error;
	// Otherwise copy response data from packet
	else
		memcpy(pData, Buff+PACKET_HEADER_SIZE+4, DataLen);

	return HEIE_NULL;
}
