/****************************************************************
 *
 * smsServCommon
 * Author:  Damon Spayde (spayde@jlab.org)
 * Time-stamp: <2003-07-30 14:50:14 spayde>
 * 
 ***************************************************************/

#include <string.h>
#include <ctype.h>
#include "smsServCommon.h"

/***************************************************************
 * Subtract the `struct timeval' values X and Y, storing the result in
 * RESULT.  Return 1 if the difference is negative, otherwise 0.
 *
 * Function copied from libc info page
 **************************************************************/
int timeval_subtract(struct timeval* result, struct timeval x, 
		     struct timeval y){
    /* Perform the carry for the later subtraction by updating Y. */
    if (x.tv_usec < y.tv_usec) {
	int nsec = (y.tv_usec - x.tv_usec) / 1000000 + 1;
	y.tv_usec -= 1000000 * nsec;
	y.tv_sec += nsec;
    }
    if (x.tv_usec - y.tv_usec > 1000000) {
	int nsec = (x.tv_usec - y.tv_usec) / 1000000;
	y.tv_usec += 1000000 * nsec;
	y.tv_sec -= nsec;
    }
    
    /* Compute the time remaining to wait.
       `tv_usec' is certainly positive. */
    result->tv_sec = x.tv_sec - y.tv_sec;
    result->tv_usec = x.tv_usec - y.tv_usec;
    
    /* Return 1 if result is negative. */
    return x.tv_sec < y.tv_sec;
}


/***************************************************************
 * This function reads a line of input from file and returns the
 * length.  Chops off the newline character.
 **************************************************************/
int getline(char *line, int max, FILE* file) {
    if (fgets(line, max, file) == NULL)
	return 0;
    else {
	char* char_ptr;
	char_ptr = strpbrk(line, "\r\n");
	*char_ptr = '\0';
	return strlen(line);
    }
}

/***************************************************************
 * This function returns a pointer to the value of a key-value pair
 * separated by an equals sign and one or more spaces iff key matches
 * second argument.
 * Return = NULL if not a match
 * Return = pointer string if match found
 **************************************************************/
char* key_value(char* ini_line, const char* key) {
    
    char* char_ptr;
    if (strncmp(ini_line, key, strlen(key)) == 0) {
	char_ptr = strpbrk(ini_line, "=");
	char_ptr++;
	if (!isalnum(*char_ptr)) {
	    char_ptr++;
	}
	return char_ptr;
    } else {
	return NULL;
    }
}
	
/***************************************************************
 * This function positions the file pointer on the line containing the
 * specified section header.
 * Return = 0 if section header founder
 * Return = -1 if reach EOF without finding section header
 **************************************************************/
int find_section(FILE* file, const char* section) {

    int length = 100;
    char ini_line[length];
    char section_head[strlen(section)+3];
    int size = 0;

    sprintf(section_head, "%c%s%c", kSectionOpen, section, kSectionClose);

    while (1) { /* Loop for section header */
	
	/* Read in line of file */
	size = getline(ini_line, length, file);
	
	/* Check for end of file */
	if ( feof(file) ) {
	    fprintf(stderr, "**find_section => Unable to find %s.\n", 
		    section_head);
	    return -1;
	}
	
	/* Check to see if the line contains the section
	   header */
	if (strncmp(ini_line, section_head, 
		    strlen(section_head)) == 0) {
	    return 0;
	}
	
    } /* End section header loop */

}
